
###################
# Class Definitions
###################

class EditBuffer():
    """
    Container for a buffer that uses a linked
    list of Python lists for sorting text
    """
    __slots__ = ('curLine', 'lst', 'curCol')

class Lst():
    """
    Container for a list and size
    """
    __slots__ = ('head', 'size')

class Empty():
    """
    Creates Empty node
    """
    __slots__ = ()

class NonEmpty():
    """
    Creates NonEmpty node
    """
    __slots__ = ('next', 'head')

class Command():
    """
    Creates Command node
    """
    __slots__ = ('cmd', 'value')

################
# Class Builders
################

def mkBuffer(lst):
    """Builder: mkBuffer: (lst) => lst => EditBuffer"""
    buff = EditBuffer()
    buff.curLine = 0
    buff.curCol = 0
    buff.lst = mkList(lst)
    return buff

def mkList(lst):
    """Builder: mkList: lst => line in buffer"""
    node = Lst()
    if lst == [] or isinstance(lst, Empty):
        node.head = mkEmpty()
        node.size = 0
        return node
    else:
        node.head = mkNonEmpty(lst, Empty())
        node.size = 1
        return node

def mkNonEmpty(head, lst):
    """Builder: mkNonEmpty: node => NonEmpty"""
    node = NonEmpty()
    node.head = head
    node.next = lst
    return node

def mkEmpty():
    """Builder: Node => Empty"""
    return Empty()

def mkCommand(cmd, value):
    """Builder: list => Command"""
    node = Command()
    node.cmd = cmd
    node.value = value
    return node

###########
# Functions
###########

def getElement(buff, line):
    """Returns buffer at specific line
    buffer + integer => list"""
    cur = buff.lst.head
    for i in range(line):
        cur = cur.next
    return cur.head

def printB( buff ):
    """Prints buffer
       NOTE: {} represents location of buffer
       buffer => output"""
    fill(buff, '{}')
    print()
    for nextLine in range(buff.lst.size):
        line = getElement(buff, nextLine)
        text = ""
        for i in range(len(line)):
            text = text + line[i]
        print(text, end="\n")
    deleteSequence(buff, 2)
    print()

def remove(buff, curLine):
    """Removes buffer at curLine
    buffer + integer => none
    post condition => line removed"""

    buff.lst.size -= 1

    if curLine == 0:
        buff.lst.head = buff.lst.head.next
    else:
        prev = buff.lst.head
        for i in range(curLine-1):
            prev = prev.next
        prev.next = prev.next.next
    if buff.curLine > 0:
        buff.curLine -= 1

def insert( buff, lst, pos ):
    """Insers line at a specific line
    buffer + list + integer => none
    post condition => line inserted"""
    
    newNode = mkNonEmpty(lst, mkEmpty())
    buff.lst.size += 1
    if isinstance( buff.lst.head, Empty ):
        buff.lst.head = newNode
    elif isinstance(buff.lst.head, NonEmpty):
        if pos == 0:
            newNode.next = buff.lst.head
            buff.lst.head = newNode
        elif pos > 0:
            cur = buff.lst.head
            position = 0
            while (position < pos-1) and (not isinstance(cur.next, Empty)):
                cur = cur.next
                position += 1
            newNode.next = cur.next
            cur.next = newNode

def forward(buff, num):
    """Moves forward by number of characters
    buffer + integer => none
    post condition => buffer moved forward by integer"""
    cur = buff.lst.head
    for i in range(buff.curLine):
        cur = cur.next
    while num > 0:
        if buff.curLine == buff.lst.size-1 and buff.curCol == len(cur.head):
            return
        if buff.curCol == len(cur.head):
            num -= 1
            if isinstance(cur.next, NonEmpty):
                buff.curLine += 1
                buff.curCol = 0
                cur = cur.next
        else:
            buff.curCol += 1
            num -= 1

def back(buff, num):
    """Moves backward by number of characters
    buffer + integer => none
    post condition => buffer moved back by integer"""
    cur = buff.lst.head
    for i in range(buff.curLine-1):
        cur = cur.next
    while num > 0:
        if buff.curCol == 0:
            if buff.curLine == 0:
                return
            num -= 1
            buff.curLine -= 1
            cur = buff.lst.head
            for i in range(buff.curLine):
                cur = cur.next
            buff.curCol = len(cur.head)
        else:
            buff.curCol -= 1
            num -= 1

def readFile(fileName):
    """Opens string from fileName and stores it in the buffer
    file name => Buffer object"""
    lst = []
    buff = mkBuffer([])
    for line in open(fileName):
        lineS = line.strip()
        for i in range(len(lineS)):
            if lineS[i] != '$':
                lst.append(lineS[i])
            else:
                insert(buff, lst, buff.lst.size)
                lst = []
    insert(buff, lst, buff.lst.size)
    lst = []
    return buff

def breakLine(buff):
    """Breaks the line into two pieces at the cursor position
    buffer => none
    post condition => line broken at buffer"""
    line = getElement(buff, buff.curLine)
    remove(buff, buff.curLine)
    col = buff.curCol

    left = line[:col]
    right = line[col:]

    insert(buff, left, buff.curLine)

    if isinstance(right, list):
        buff.curLine += 1
        insert(buff, right, buff.curLine)
        buff.curCol = 0

def insertChar(buff, char):
    """Inserts character into buffer at cursor location
    buffer + character => none
    post condition => character inserted at buffer"""
    if char == '$' :
        breakLine(buff)
    elif char == '_':
        cur = buff.lst.head
        for i in range(buff.curLine):
            cur = cur.next
        cur.head.insert(buff.curCol, ' ')
        buff.curCol += 1
    else:
        cur = buff.lst.head
        for i in range(buff.curLine):
            cur = cur.next
        cur.head.insert(buff.curCol, char)
        buff.curCol += 1

def deleteChar(buff):
    """Deletes character in buffer at cursor location
    buffer => none
    post condition => character deleted at buffer"""
    cur = buff.lst.head
    if buff.lst.head.head == []:
        return
    if buff.curLine == 0:
        buff.lst.head.head.pop(buff.curCol-1)
        if buff.curCol > 0:
            buff.curCol -=1
        else:
            buff.curCol = 0
    elif buff.curLine == 2:
        if buff.curCol == 0:
            for i in range(buff.curLine-1):
                cur = cur.next
            left = cur.head
            right = cur.next.head
            pos = len(left)
            lst = left + right
            remove(buff, buff.curLine)
            remove(buff, buff.curLine)
            insert(buff, lst, buff.curLine+1)
            buff.curLine += 1
            buff.curCol = pos
            return
        for i in range(buff.curLine):
            cur = cur.next
        cur.head.pop(buff.curCol-1)
        if buff.curCol > 0:
            buff.curCol -=1
        else:
            buff.curCol -= 1
    else:
        if buff.curCol == 0:
            for i in range(buff.curLine-1):
                cur = cur.next
            left = cur.head
            right = cur.next.head
            pos = len(left)
            lst = left + right
            remove(buff, buff.curLine)
            remove(buff, buff.curLine)
            if buff.curLine == 0:
                insert(buff, lst, 0)
            else:
                insert(buff, lst, buff.curLine+1)
                buff.curLine += 1
            buff.curCol = pos
            return
        for i in range(buff.curLine):
            cur = cur.next
        cur.head.pop(buff.curCol-1)
        if buff.curCol > 0:
            buff.curCol -=1
        else:
            buff.curCol -= 1

def fill(buff, string):
    """Inserts string of characters into buffer at cursor location
    buffer + string => none"""
    for i in range(len(string)):
        insertChar(buff, string[i])

def deleteSequence(buff, num):
    """Deletes sequence of string with num being how many
       characters to delete
       buffer + number => none"""
    for i in range(num):
        deleteChar(buff)

def takeCommand():
    """Asks for input, takes command, returns object
    none => Command object"""
    command = input("Enter command: ")
    command = command.split()
    if len(command) == 1:
        return mkCommand(command[0], mkEmpty())
    elif len(command) == 2:
        return mkCommand(command[0], command[1])
    else:
        print("Input error, retry...")
        takeCommand()

############
# Test Cases
############

def main():
    """Main function"""
    fileName = input("Enter file name: ")
    buff = readFile(fileName)
    print("\nCommands:\n\nforward *integer*  => moves buffer forward\nback *integer*     => moves buffer back\ninsert *char*      => inserts character\ndelete *integer*   => deletes n number of characters\nfill *string*      => inserts string ('_' detonates space)\nhelp               => command list\nexit               => enters raw shell\n***Note, leaving any of the integer variables blank will result in a value of 1***")
    input("\nTo confirm, press enter")
    printB(buff)

    while True:
        ch = takeCommand()
        if ch.cmd == 'forward':                 # FORWARD
            if isinstance(ch.value, Empty):
                forward(buff, 1)
            else:
                forward(buff, int(ch.value))
            printB(buff)
        elif ch.cmd == 'back':                  # BACK
            if isinstance(ch.value, Empty):
                back(buff, 1)
            else:
                back(buff, int(ch.value))
            printB(buff)
        elif ch.cmd == 'insert':                # INSERT
            if isinstance(ch.value, Empty):
                print("command not recognized")
            else:
                insertChar(buff, ch.value)
            printB(buff)
        elif ch.cmd == 'delete':                # DELETE
            if isinstance(ch.value, Empty):
                deleteChar(buff)
            else:
                deleteSequence(buff, int(ch.value))
            printB(buff)
        elif ch.cmd == 'fill':                  # FILL
            if isinstance(ch.value, Empty):
                print("command not recognized")
            else:
                fill(buff, ch.value)
            printB(buff)
        elif ch.cmd == 'help':                  # HELP
            print("\nCommands:\n\nforward *integer*  => moves buffer forward\nback *integer*     => moves buffer back\ninsert *char*      => inserts character\ndelete *integer*   => deletes n number of characters\nfill *string*      => inserts string ('_' detonates space)\nhelp               => command list\nexit               => enters raw shell\n***Note, leaving any of the integer variables blank will result in a value of 1***\n")
            
        elif ch.cmd == 'exit':                  # EXIT
            print("Goodbye...\n")
            return
        
        else:
            print("command not recognized")

main()
